#include "util/version.hpp"
namespace horizon {
const unsigned int Version::major = 0;
const unsigned int Version::minor = 9;
const unsigned int Version::micro = 0;
const char *Version::name = "Aurora";
const char *Version::commit = "682de80 2019-06-17 01:26:16 +0200 imp: zoom to fit on open";
}
